####################################
### Load libraries and functions ###
####################################
library(evd)
library(Rcpp)
library(RcppArmadillo)
library(inline)
sourceCpp("Functions/likelihood.cpp")

### Function that simulates n alpha-stable random variates (see Stephenson, 2003)
# n: integer, number of random variates
# alpha: real number, parameter
rstable <- function(n,alpha){
	U <- runif(n)
	V <- runif(n)
	W <- -log(V) ### standard exponential variables
	return( (sin((1-alpha)*U*pi)/W)^((1-alpha)/alpha)*sin(alpha*U*pi)/sin(U*pi)^(1/alpha) )	
}

### Function that simulates n d-dimensional extreme-value nested logistic random variates with two layers (see Stephenson, 2003, p.58 and Algorithm 2.1-2.2)
# n: integer, number of logistic random vectors
# d: integer, dimension of the random vector
# alphas: vector of real numbers, alphas[1] is the main dependence parameter (alpha) while alphas[2:end] is the vector of dependence parameters for sub-groups (alpha.b)
# groups: list of vectors of integers, containing the indices of variables for each sub-group
rnlogistic <- function(n,d,alphas,groups){
	alpha <- alphas[1]
	alpha.b <- alphas[2:length(alphas)]
	
	S <- matrix(rstable(n,alpha),nrow=n,ncol=d) ### underlying stable variates for layer 1
	Sb <- matrix(nrow=n,ncol=d) ### underlying stable variates for layer 2
	alb <- matrix(nrow=n,ncol=d) ### dependence parameters for layer 2 in matrix form 
	for(b in 1:length(groups)){
		Sb.vec <- rstable(n,alpha.b[b])
		Sb[,groups[[b]]] <- Sb.vec
		alb[,groups[[b]]] <- alpha.b[b]
	}
	
	W <- matrix(-log(runif(n*d)),nrow=n,ncol=d) ### standard exponential variates
	return( (Sb*S^(1/alb)/W)^(alpha*alb) )
}


##############
### diag.plot ###
##############
#diag.plot plots MCMC output
#Input:
#burnin: numeric value indicating burn it period
#out.MCMC: MCMC function output
#Output:
#plot MCMC output
diag.plot <- function(out.MCMC) {
	burnin=out.MCMC$burnin
	R=nrow(out.MCMC$chain)
	K<-length(unique(out.MCMC$groups))
	par(mfrow = c((K+1), 2))
    plot(out.MCMC$out.alphas.chain[, 1], xlab = "Iterations", ylab = expression(alpha), type = "b", pch = 20, main = "MCMC output", ylim = c(0, 1))
    abline(h = out.MCMC$alphas[1], col = "red")
    text(x = R, y = 0, labels = paste("Acceptance rate: ", 100 * round(out.MCMC$accept.rate.alphas[1], digits = 2), "%", sep = ""), pos = 2, col = "purple")
    acf(out.MCMC$out.alphas.chain[1:R, 1], main = "ACF of MCMC output after burn-in")
   
	for(k in 2:(K+1)){
        plot(out.MCMC$out.alphas.chain[, 1]*out.MCMC$out.alphas.chain[, k], xlab = "Iterations", ylab = expression(alpha * alpha[k]), type = "b", pch = 20, ylim = c(0, 1))
        abline(h = out.MCMC$alphas[1]*out.MCMC$alphas[k], col = "red")
        text(x = R, y = 0, labels = paste("Acceptance rate: ", 100 * round(out.MCMC$accept.rate.alphas[k], digits = 2), "%", sep = ""), pos = 2, col = "purple")
        acf(out.MCMC$out.alphas.chain[1:R, k], main = "")
	}
}

# If alpha is equal to 1 then we split the elements inside alpha in different groups 
update.groups<- function(alphas.b.current, Groups.current, D_K.current, eps.b, K, K.max){
	if (K==1){
		K.new=K.max
		D_K.new<-rep(1, D)
		Groups.new<-c(1:D)
		alphas.b.new<-rep(alphas.b.current, K.new)
		eps.b.new<-rep(eps.b, K.new)
	}else{
		D_K.new<- D_K.current
		eps.b.current<-eps.b
		eps.b.new<-eps.b
		Groups.new<-Groups.current
		Groups.current.new<-Groups.current
		alphas.b.new<-alphas.b.current
		K.new<-K
		for (k in 1:K) {
			if (alphas.b.current[k]==1 & D_K.current[k]>1){ #if alpha_b is equal to 1 and D_K is different from 1 then split group in D_K subgroups and assign value 1
				G<-which(Groups.current==k)
				D_K.new[Groups.new[G[1]]]<-1
				for (i in 2:length(G)) {
					value<-max(Groups.new[1:(G[i])]) # max of groups until the first variable belonging to the second cluster of the split
#consider the case where alpha is 1!!!
					y<-which(Groups.new>value)
					Groups.new[y]<-Groups.new[y]+1
					Groups.new[G[i]]<-value+1
					if (value==K.new){
						alphas.b.new<-c(alphas.b.new[1:value], alphas.b.current[k])
						eps.b.new<-c(eps.b.new[1:value], eps.b.current[k])
						D_K.new<-c(D_K.new[1:value], 1)
					}else{
						alphas.b.new<-c(alphas.b.new[1:value], alphas.b.current[k], alphas.b.new[(value+1):K.new])
						eps.b.new<-c(eps.b.new[1:value], eps.b.current[k], eps.b.new[(value+1):K.new])
						D_K.new<-c(D_K.new[1:value], 1, D_K.new[(value+1):K.new])
					}
					K.new<-K.new+1
				}
			}
		}
		K.new<-length(unique(Groups.new))
	}
	return(list(alphas.b.new=alphas.b.new, D_K.new=D_K.new, Groups.new=Groups.new, K.new=K.new, eps.b=eps.b.new))
}


split.update.groups<- function(p, u, alphas.b.current, Groups.current, D_K.current, eps.b.current, K){
    d_k<-sample(1:(D_K.current[p]-1), 1)
    Groups.new<-Groups.current
    t<-which(Groups.current==p)
    G<-sample(1:length(t), d_k) # Need to match D_K and Groups order
    if (min(t[G])<min(t[-G])){
       G1<-t[G]
       G2<-t[-G]
    }else{
        G1<-t[-G]
        G2<-t[G]
    }
    Groups.new[G1]<-p
    value<-max(Groups.current[1:(min(G2))]) # max of groups until the first variable belonging to the second cluster of the split
    Groups.new[G2]<-value+1
#consider the case where alpha is 1!!!
    y<-which(Groups.current>value)
    Groups.new[y]<-Groups.new[y]+1
    if (value==K){
        alphas.b.new<-c(alphas.b.current[1:value], alphas.b.current[p]-u)
        eps.b.new<-c(eps.b.current[1:value], eps.b.current[p])
        D_K.new<-c(D_K.current[1:value], length(G2))
    }else{
        alphas.b.new<-c(alphas.b.current[1:value], alphas.b.current[p]-u, alphas.b.current[(value+1):K])
        eps.b.new<-c(eps.b.current[1:value], eps.b.current[p], eps.b.current[(value+1):K])
        D_K.new<-c(D_K.current[1:value], length(G2), D_K.current[(value+1):K])
    }
    alphas.b.new[p]<-alphas.b.current[p]+u # this is a bit nasty but the system of equation does not have a coomon solution so not sure howelse I could do
    eps.b.new[p]<-eps.b.current[p]
    D_K.new[p]<-length(G1)
    K.new<-length(unique(Groups.new))
    return(list(alphas.b.new=alphas.b.new, D_K.new=D_K.new, Groups.new=Groups.new, K.new=K.new, eps.b.new=eps.b.new))
}

swap.update.groups<- function(p, c, alphas.b.current, Groups.current, D_K.current, eps.b.current, K){
    d_k<-sample(1:(D_K.current[p]-1), 1)
    Groups.new<-Groups.current
    t<-which(Groups.current==p)
    G<-sample(1:length(t), d_k) # Need to match D_K and Groups order
    if (min(t[G])<min(t[-G])){
       G1<-t[G]
       G2<-t[-G]
    }else{
        G1<-t[-G]
        G2<-t[G]
    }
    Groups.new[G1]<-p
        D_K.new<-D_K.current
    if(p>c | min(G2)>min(which(Groups.current==c))){
     	Groups.new[G2]<-c
     	alphas.b.new<-c(alphas.b.current)
     	eps.b.new<-c(eps.b.current)
     	D_K.new<-D_K.current
    	D_K.new[p]<-length(G1)
    	D_K.new[c]<-(D_K.new[c]+length(G2))
    }else{
    	 value<-max(Groups.current[1:(min(G2))]) # max of groups until the first variable belonging to the second cluster of the split
     	Groups.new[G2]<-value+1
     	y<-which(Groups.current>value & Groups.current<(c+1))
    	Groups.new[y]<-Groups.new[y]+1
    	z<-which(Groups.current==c)
    	Groups.new[z]=value+1
    	D_K.c<-length(G2)+length(z)
    	D_K.current<-D_K.current[-c]
    	alpha.c<-alphas.b.current[c]
    	alphas.b.current<-alphas.b.current[-c]
    	eps.c<-eps.b.current[c]
		eps.b.current<-eps.b.current[-c]
    	if (value==(K-1)){
    		alphas.b.new<-c(alphas.b.current[1:value], alpha.c)
    		eps.b.new<-c(eps.b.current[1:value], eps.c)
        	D_K.new<-c(D_K.current[1:value], D_K.c)
    	}else{
    		alphas.b.new<-c(alphas.b.current[1:value], alpha.c, alphas.b.current[(value+1):(K-1)])
    			eps.b.new<-c(eps.b.current[1:value], eps.c, eps.b.current[(value+1):(K-1)])
        	D_K.new<-c(D_K.current[1:value], D_K.c, D_K.current[(value+1):(K-1)])
    	}
    }
    D_K.new[p]<-length(G1)
     return(list(alphas.b.new=alphas.b.new, D_K.new=D_K.new, Groups.new=Groups.new, eps.b.new=eps.b.new))
}

merge.update.groups<- function(p, alphas.b.current, Groups.current, D_K.current, eps.b.current, K){
    alphas.b.new<-alphas.b.current
    alphas.b.new[p[1]]<-(alphas.b.current[p[1]]+alphas.b.current[p[2]])/2
    alphas.b.new<-alphas.b.new[-p[2]]
    eps.b.new<-eps.b.current
    eps.b.new[p[1]]<-(eps.b.current[p[1]]+eps.b.current[p[2]])/2
    eps.b.new<-eps.b.new[-p[2]]
    #update D_K and Groups to estimate log ratio
    D_K.new<-D_K.current
    D_K.new[p[1]]<-(D_K.new[p[1]]+D_K.new[p[2]]) # we incorporate group 2 in group 1, maybe better way to do it
    D_K.new<- D_K.new[-p[2]]
    Groups.new<-Groups.current 
    t<-which(Groups.current==p[2])
    Groups.new[t]<-p[1]
    y<-which(Groups.current>p[2])
    Groups.new[y]<-Groups.new[y]-1
    K.new<-length(unique(Groups.new))
    return(list(alphas.b.new=alphas.b.new, D_K.new=D_K.new, Groups.new=Groups.new, K.new=K.new, eps.b.new=eps.b.new))
}

##############
### RJMCMC ###
##############
### Performs RJ-MCMC algorithm using the recursive formula of the nested logistic likelihood
# data: numeric matrix containing the extreme observations in Fréchet scale
# D: integer, data dimension
# R: number of iterations
# init.Groups: integer vector containing initial clustering configuration, same integer indicate that the corresponding variables belong to the same group
# init.D_K: integer vector containing the size of each initial group
# init.alphas.a: numeric, contains the initial value for the between clusters dependence parameter
# init.alphas.b: numeric vector, contains the initial value for the within clusters dependence parameter
# eps.init: adaptive tuning parameter for the proposal in the UPDATE move
# eta: arbitrary constant for the proposal in the SPLIT and MERGE moves, in practice should be chosen in order to ensure the RJ MCMC chain is well-mixing. We reccomend fixing it to 0.4.
RJMCMC <- function(data=na.omit(MAX), D=4, R = 1000, init.Groups =c(1,1,1,2), burnin=R/5, init.D_K=c(3,1), init.alphas.a = c(0.8), init.alphas.b = c(0.4, 0.8), eps.init = c(0.1, 0.1, 0.1), eta=0.4) { # eta too small means not much merging
	# need to think about how to define initial values for the parameters and tree structure
	n <- nrow(data) # sample size
	eps=eps.init
	K.max = D # there can be as many group as the dimensions, would be nice to make K.max flexible, i.e., K.max<=D
	out.K <- c() # chain for number of groups
	K<-out.K[1] <- length(init.alphas.b) # number of groups
	out.alphas.a <- c() # chain for alpha parameters
	out.alphas.a[1] <- init.alphas.a
	out.alphas.b <- matrix(NA, (R+1), K.max) # chain for alphas.b parameters
	out.alphas.b[1, 1:K] <- init.alphas.b
	D_K <- matrix(NA, (R+1), K.max) # chain for D_K
	D_K[1,1:K]<-init.D_K
	Groups <- matrix(NA, (R+1), D)
	Groups[1,] <- init.Groups # ordered depending on the apperance of the group in the columns of the data matrix 
	log.post <- c()
	log.post[1]  <- -NegLikelihoodnestedLogistic2(c(init.alphas.a,  init.alphas.b), data, init.Groups, init.D_K)
	accept.alphas <- matrix(, ncol = (K.max+1), nrow = R)
	accept.tree <- c()
	accept.tree.split <- c()
	accept.tree.merge <- c()
	accept.tree.swap<- c()
	for (r in 1:R) {# main loop
		K=out.K[r]
		alphas.b.current=out.alphas.b[r, 1:K]
		Groups.current=Groups[r,]
		D_K.current=D_K[r, 1:K]
		eps.b=eps[2:(K+1)]
		UPDATE<-update.groups(alphas.b.current, Groups.current, D_K.current, eps.b, K, K.max)
		K<-UPDATE$K.new
		eps<-c(eps[1], UPDATE$eps.b)
		D_K[r, 1:K]<-UPDATE$D_K.new	
		out.alphas.b[r, 1:K]<-UPDATE$alphas.b.new
		Groups[r,]<-UPDATE$Groups.new
		out.K[r]<-K
		Groups.current<-Groups[r,]
		D_K.current<-D_K[r, 1:K]
		if (runif(1) < 0.5) {
			###############
			### UPDATE ###
			###############
			alphas.current<-c(out.alphas.a[r], out.alphas.b[r, 1:K])
			Groups[r+1,] <- Groups.current
			D_K[r+1, 1:K] <- D_K.current
			out.K[r+1]<-K
			log.post.current<-log.post[r]
			accept.alphas.current <- c()
			D_K.current.ab<-c(K, D_K.current) # we only care about the k group so the alpha parameter has k elements but doesn't matter
			for (p in 1:(K+1)) { #Loop over the parameters to estimate	
				#if (D_K.current.ab[p]==1){
				#	alphas.current <- alphas.current
				#	log.post.current  <- log.post.current
				#	accept.alphas.current [p]<- NA
				#}else{
					if ((alphas.current[p] != 1) & (alphas.current[p] - eps[p]) < 1 & (alphas.current[p] + eps[p]) >= 1 & K!=K.max) { # If 1 is in the inverval but alpha_b is not exactly 1
						 # indicator for proposal distribution
						if (runif(1) < 0.5) {
							pr = 1
							alpha.star <- runif(1, min = max(0.1, alphas.current[p] - eps[p]), max = min(1, alphas.current[p] + eps[p]))
						} else {
							pr = 2
							alpha.star <- 1 # Point mass at 1 with probability 0.5
						}
					} else {
						pr = 0
						alpha.star <- runif(1, min = max(0.1, alphas.current[p] - eps[p]), max = min(1, alphas.current[p] + eps[p])) ## generation of a new parameter centred around the current parameter (bouded at 0.1 and 1)
					}
					alphas.star <- alphas.current
					alphas.star[p] <- alpha.star
					log.post.star <- -NegLikelihoodnestedLogistic2(alphas.star, data, Groups.current, D_K.current)
					if (alpha.star == 1) { #the computation of the log probabilities changes depending from which proposal the candidate was drawn from
						p_alpha_star=0.5+ 0.5*dunif(alpha.star, min = max(0.1, alphas.current[p] - eps[p]), max = min(1, alphas.current[p] + eps[p]), log = TRUE)
					}else{
						if ((alphas.current[p] - eps[p]) < 1 & (alphas.current[p] + eps[p]) >= 1 & K!=K.max) { 
							p_alpha_star=0.5*dunif(alpha.star, min = max(0.1, alphas.current[p] - eps[p]), max = min(1, alphas.current[p] + eps[p]), log = TRUE)
						}else{
							p_alpha_star=dunif(alpha.star, min = max(0.1, alphas.current[p] - eps[p]), max = min(1, alphas.current[p] + eps[p]), log = TRUE)
						}
					}
					if (alphas.current[p] == 1) { # the computation of the log probabilities changes depending from which proposal the candidate was drawn from
						p_alpha_current=0.5+ 0.5*dunif(alphas.current[p], min = max(0.1, alpha.star - eps[p]), max = min(1, alpha.star + eps[p]), log = TRUE)
					}else{
						if ((alpha.star - eps[p]) < 1 & (alpha.star + eps[p]) >= 1 & K!=K.max) { 
							p_alpha_current=0.5*dunif(alphas.current[p], min = max(0.1, alpha.star - eps[p]), max = min(1, alpha.star + eps[p]), log = TRUE)
						}else{
							p_alpha_current=dunif(alphas.current[p], min = max(0.1, alpha.star - eps[p]), max = min(1, alpha.star + eps[p]), log = TRUE)
						}
					}
					log.prob <- min(0, log.post.star - log.post.current + p_alpha_star -  p_alpha_current) ## calculation of the log-acceptance probability
					if ((runif(1) < exp(log.prob)) & (alphas.current[p] != alpha.star)) { ## with probability exp(log.prob), we accept the new parameter, and update the chain accordingly, if the previous parameter value is the same as the candidate it is not count as an acceptance (suggested by Segers)
						alphas.current <- alphas.star
						log.post.current  <- log.post.star
						accept.alphas[r, p]<- 1
					} else { ## with probability 1-exp(log.prob), we stay at the current parameter value, and update the chain accordingly
						alphas.current <- alphas.current
						log.post.current  <- log.post.current
						accept.alphas[r, p]<- 0
					}
					# Adaptive steps for epsilon 
					if ((r<=burnin) & (r%%100==0) & (r>0)) {
					if ((D_K.current.ab[p]>1) & (eps[p] + 0.05 < 0.2) & (mean(na.omit(accept.alphas[, p])) > 0.4)) { 
					eps[p] <- eps[p] + 0.05
					} else {
						if ((D_K.current.ab[p]>1) & (eps[p] - 0.05 > 0.05) & (mean(na.omit(accept.alphas[, p])) < 0.2)) {
							eps[p] <- eps[p] - 0.05
						}
					}
				}
			}
			out.alphas.a[r + 1] <- alphas.current[1]
			out.alphas.b[r + 1, 1:K]<-alphas.current[-1]
			log.post[r + 1] <- log.post.current
		}else{
		############
		### SPLIT ###
		############
			if (runif(1) < 0.33 + 0.33 * (K == 1) - 0.33 * (K == K.max)) { # Split: k->k+1, gives more prob if k=1 and less if k is already big
				#D_K? once we decide to add one group we need to identify which variables will reside in this group, one way could be the following:
				alphas.a.current <- out.alphas.a[r]
				alphas.b.current <- out.alphas.b[r, 1:K]
				# Selection procedure for group to split 
				select<-c()
				for (g in 1:K){
					if (D_K.current[g]>1 & alphas.b.current[g]<0.99){ # Make sure we sample from groups with at least 2 variables for the split move to make sense
						select[g]<-(alphas.b.current[g]) # Important to tune weel the weights
					}else{
						select[g]<-0
					}
				}
				if(any(select!=0)){
					p<-sample(1:K, 1, prob=((select/sum(select))*(1/K)))
					u<-runif(1, min=max((-min(1,(alphas.b.current[p]+eta))+alphas.b.current[p]), (max(0.1, (alphas.b.current[p]-eta))- alphas.b.current[p])), max=min((-max(0.1,(alphas.b.current[p]-eta))+alphas.b.current[p]), (min(1, (alphas.b.current[p]+eta))- alphas.b.current[p])))
	     			# update alphas, D_K and Groups to estimate log ratio 
					SPLIT.UPDATE<-split.update.groups(p=p, u=u,  alphas.b.current=out.alphas.b[r, 1:K], Groups.current=Groups[r,], D_K.current=D_K[r, 1:K], eps.b.current=eps[2:(K+1)], K=K)
					alphas.b.new=SPLIT.UPDATE$alphas.b.new
					D_K.new=SPLIT.UPDATE$D_K.new
					Groups.new=SPLIT.UPDATE$Groups.new
					K.new<-SPLIT.UPDATE$K.new
					eps.new<-c(eps[1], SPLIT.UPDATE$eps.b.new)
					Jacobian<- log(2)
					log.post.star <- -NegLikelihoodnestedLogistic2(c(alphas.a.current, alphas.b.new), data, Groups.new, D_K.new)
						log.prob <- min(0, (log.post.star - log.post[r] - dunif(u, min=max((-min(1,(alphas.b.current[p]+eta))+alphas.b.current[p]), (max(0.1, (alphas.b.current[p]-eta))- alphas.b.current[p])), max=min((-max(0.1,(alphas.b.current[p]-eta))+alphas.b.current[p]), (min(1, (alphas.b.current[p]+eta))- alphas.b.current[p])), log=TRUE) + Jacobian ))
					if (runif(1) < exp(log.prob)) {
						K<-out.K[r+1] <-K.new
						out.alphas.a[r+1] <- out.alphas.a[r]
						out.alphas.b[r+1,1:K] <- alphas.b.new
						D_K[r+1, 1:K]<-D_K.new
						Groups[r+1,] <- Groups.new
						log.post[r + 1] <- log.post.star
					#	accept.tree[r] <- 1
						accept.tree.split[r] <- 1
						eps<-eps.new
					} else {
						out.alphas.a[r+1] <- out.alphas.a[r ]
						out.alphas.b[r+1,] <- out.alphas.b[r,] 
						out.K[r+1] <- K 
						Groups[r+1,] <- Groups[r,]
						D_K[r+1, ]<-D_K[r, ]
						log.post[r + 1] <- log.post[r]
					#	accept.tree[r] <- 1
						accept.tree.split[r] <- 0
					}
				}else{
					out.alphas.a[r+1] <- out.alphas.a[r]
					out.alphas.b[r + 1,] <- out.alphas.b[r,] 
					out.K[r + 1] <- K 
					Groups[r + 1,] <- Groups[r,]
					D_K[r + 1, ]<-D_K[r, ]
					log.post[r + 1] <- log.post[r]
				#	accept.tree[r] <- 0
				} # end loop if close enough we can do the split
				#rate.accept.tree <- mean(na.omit(accept.tree)) #
				#if ((eta - 0.05 > 0.05) & (rate.accept.tree.split > 0.4)) { # Adaptive steps for eta 
				#	eta <- eta - 0.05
				#} 
			}else{ # Merge: k->k-1
				if (runif(1) < 0.33 - 0.33 * (K == 1) + 0.33 * (K == K.max)){
					##############
					### MERGE  ###
					##############
					alphas.a.current <- out.alphas.a[r]
					alphas.b.current <- out.alphas.b[r, 1:K]
					select<-c()
					if (K>1){
						p_select<-c()
		 				for (l in 1:(K-1)){ 
		 				for (m in (l+1):K){
		 					p_try<-c(l, m)
		 					if ((max(0.1, (alphas.b.current[p_try[1]]-eta))- alphas.b.current[p_try[1]])<(alphas.b.current[p_try[1]]-alphas.b.current[p_try[2]]) & (alphas.b.current[p_try[1]]-alphas.b.current[p_try[2]])<(min(1, (alphas.b.current[p_try[1]]+eta))-alphas.b.current[p_try[1]])){ # we can only join groups with alpha parameters close to each other (this constraint comes from the boundary we choose to simulated the u in the split move)
		 						p_select<-rbind(p_select, p_try)
		 						index<-((alphas.b.current[p_try[1]]*(1/2))+(alphas.b.current[p_try[2]]*(1/2))) # Important to tune weel the weights
		 						select<-c(select, index)
		 					}
		 				}
					}
				}
		 		if(class(select)!="NULL"){
		 			p_sample<-sample(1:nrow(p_select), 1, prob=((select/sum(select))*(1/nrow(p_select))))
		 			p<-p_select[p_sample,]
		 			MERGE.UPDATE<-merge.update.groups(p=p, alphas.b.current=out.alphas.b[r, 1:K], Groups.current=Groups[r,], D_K.current=D_K[r, 1:K], eps.b.current=eps[2:(K+1)], K=K)
					alphas.b.new=MERGE.UPDATE$alphas.b.new
					D_K.new=MERGE.UPDATE$D_K.new
					Groups.new=MERGE.UPDATE$Groups.new
					K.new<-MERGE.UPDATE$K.new
					eps.new<-c(eps[1], MERGE.UPDATE$eps.b.new)
					Jacobian<- log(1/2) 
					log.post.star <- -NegLikelihoodnestedLogistic2(c(alphas.a.current, alphas.b.new), data, Groups.new, D_K.new)
						log.prob <- min(0.1, (log.post.star - log.post[r] - dunif(alphas.b.current[p[1]]-alphas.b.new[p[1]], min=max((-min(1,(alphas.b.new[p[1]]+eta))+alphas.b.new[p[1]]), (max(0.1, (alphas.b.new[p[1]]-eta))- alphas.b.new[p[1]])), max=min((-max(0.1,(alphas.b.new[p[1]]-eta))+alphas.b.new[p[1]]), (min(1, (alphas.b.new[p[1]]+eta))- alphas.b.new[p[1]])), log=TRUE) + Jacobian )) #alphas.b.new[p[1]]=theta, alphas.b.current[p[1]]=theta_1
					if (runif(1) < exp(log.prob)) {
						K<-out.K[r+1] <-K.new
						out.alphas.a[r+1] <- out.alphas.a[r]
						out.alphas.b[r+1,1:K] <- alphas.b.new
						D_K[r+1, 1:K]<-D_K.new
						Groups[r+1,] <- Groups.new
						log.post[r + 1] <- log.post.star
						accept.tree[r] <- 1
						eps<-eps.new
						accept.tree.merge[r] <- 1
					} else {
						out.alphas.a[r+1] <- out.alphas.a[r]
						out.alphas.b[r + 1,] <- out.alphas.b[r,] 
						out.K[r + 1] <- K 
						Groups[r + 1,] <- Groups[r,]
						D_K[r + 1, ]<-D_K[r, ]
						log.post[r + 1] <- log.post[r]
						accept.tree[r] <- 1
						accept.tree.merge[r] <- 0
					}
				}else{
					out.alphas.a[r+1] <- out.alphas.a[r]
					out.alphas.b[r + 1,] <- out.alphas.b[r,] 
					out.K[r + 1] <- K 
					Groups[r + 1,] <- Groups[r,]
					D_K[r + 1, ]<-D_K[r, ]
					log.post[r + 1] <- log.post[r]
					accept.tree[r] <- 0
				} # end loop if close enough we can do the merge
					rate.accept.tree<- mean(na.omit(accept.tree)) # adaptive step for eta
					if ((eta + 0.05 <= 0.4) & (rate.accept.tree  < 0.2)) {
					eta  <- eta + 0.05
				}else{
					if ((eta - 0.05 >= 0.1) & (rate.accept.tree  > 0.4)) {
						eta  <- eta - 0.05
					}
				}
				}else{
					#############
					### SWAP  ###
					#############
					alphas.a.current <- out.alphas.a[r]
					alphas.b.current <- out.alphas.b[r, 1:K]
					select<-c()
					if (K>1){
						p_select<-c()
						for (l in 1:(K-1)){ 
		 					for (m in (l+1):K){
		 						p_try<-c(l, m)
		 						if (D_K.current[p_try[1]]>1 | D_K.current[p_try[2]]>1){
		 							p_select<-rbind(p_select, p_try)
		 							index<-((alphas.b.current[p_try[1]]*(1/2))+(alphas.b.current[p_try[2]]*(1/2))) # Important to tune well the weights
		 							select<-c(select, index)
		 						}
							}
						}
					}
		 			if(class(select)!="NULL"){
		 			p_sample<-sample(1:nrow(p_select), 1, prob=((select/sum(select))*(1/nrow(p_select))))
		 			if (D_K.current[p_select[p_sample, 1]]>1 & D_K.current[p_select[p_sample, 2]]>1){
		 				w<-sample(1:2,1)
		 				p<-p_select[p_sample, w]
		 				c<-p_select[p_sample,c(1:2)[-w]]	
		 			}else{
		 				if (D_K.current[p_select[p_sample, 1]]>1){
		 					p<-p_select[p_sample, 1]
		 					c<-p_select[p_sample, 2]
		 				}else{
		 					p<-p_select[p_sample, 2]
		 					c<-p_select[p_sample, 1]
		 				}	
		 			}
		 			SWAP.UPDATE<-swap.update.groups(p=p, c=c, alphas.b.current=out.alphas.b[r, 1:K], Groups.current=Groups[r,], D_K.current=D_K[r, 1:K], eps.b.current=eps[2:(K+1)], K=K)
					alphas.b.new=SWAP.UPDATE$alphas.b.new
					D_K.new=SWAP.UPDATE$D_K.new
					Groups.new=SWAP.UPDATE$Groups.new
					eps.new<-c(eps[1], SWAP.UPDATE$eps.b.new)
					log.post.star <- -NegLikelihoodnestedLogistic2(c(alphas.a.current, alphas.b.new), data, Groups.new, D_K.new)
						log.prob <- min(0.1, (log.post.star - log.post[r])) 
						if (runif(1) < exp(log.prob)) {
							out.K[r+1] <-K
							out.alphas.a[r+1] <- out.alphas.a[r]
							out.alphas.b[r+1,1:K] <- alphas.b.new
							D_K[r+1, 1:K]<-D_K.new
							Groups[r+1,] <- Groups.new
							log.post[r + 1] <- log.post.star
							accept.tree[r] <- 1
							eps<-eps.new
							accept.tree.swap[r] <- 1
						} else {
							out.alphas.a[r+1] <- out.alphas.a[r]
							out.alphas.b[r + 1,] <- out.alphas.b[r,] 
							out.K[r + 1] <- K 
							Groups[r + 1,] <- Groups[r,]
							D_K[r + 1, ]<-D_K[r, ]
							log.post[r + 1] <- log.post[r]
							accept.tree[r] <- 1
							accept.tree.swap[r] <- 0
						}
					}else{
						out.alphas.a[r+1] <- out.alphas.a[r]
						out.alphas.b[r + 1,] <- out.alphas.b[r,] 
						out.K[r + 1] <- K 
						Groups[r + 1,] <- Groups[r,]
						D_K[r + 1, ]<-D_K[r, ]
						log.post[r + 1] <- log.post[r]
						accept.tree[r] <- 0
					} # end loop if close enough we can do the merge
				} # end merge else loop 
			} # end else split
		} # end else jump
	} # end main loop
	chain<-cbind(out.alphas.a, out.alphas.b[1:(R+1),])
	#burnin <- floor(burnin)
	out.chain <- chain[burnin:(R+1), ] #Eliminate first observations
	out.Groups <- Groups[burnin:(R+1), ] 
	groups_distribution<-aggregate(list(numdup=rep(1,nrow(as.data.frame(out.Groups)))), as.data.frame(out.Groups), length) # computes the distribution of	
	groups <- groups_distribution[which.max(groups_distribution[, D+1]), 1:D] # Mode of Groups out of chain
	out.accept.alphas<-accept.alphas[burnin:R, ]
	out.alphas <- c()
	out.accept.alphas.rate<-c()
	for (i in 1:(R-burnin)) { #We only consider the chain values at spaced intervals for the estimation of the parameters 
		if (all(groups==out.Groups[i, ])){
			out.alphas <- rbind(out.alphas, out.chain[i, ])
			out.accept.alphas.rate <- rbind(out.accept.alphas.rate, out.accept.alphas[i, ])
		}
	}
	#thining <- as.numeric(round((nrow(out.alphas)/5), 0) - 1)
	#a = 0
	#out.al <- matrix(, ncol = ncol(chain), nrow = thining)
	#for (i in 1:thining) { #We only conder the chain values at spaced intervals for the estimation of the parameters 
		#a <- a + 5
		#out.al[i, ] <- out.alphas[a, ]
	#}
	out.al<-out.alphas
	alphas <- apply(out.al, 2, function(x) median(na.omit(x))) # Mean of parameters estimates out of thinning chain
	sd <- (apply(out.al, 2, function(x) sd(na.omit(x)))/nrow(out.alphas)) #Sd estimates
	accept.rate.tree <- mean(na.omit(accept.tree)) #
	accept.rate.tree.split <- mean(na.omit(accept.tree.split)) #
	accept.rate.tree.merge <- mean(na.omit(accept.tree.merge)) #
	accept.rate.tree.swap <- mean(na.omit(accept.tree.swap)) #
	accept.alphas.rate = apply(out.accept.alphas.rate, 2, function(x) mean(na.omit(x)))
	return(list(chain = out.chain, out.alphas.chain = out.alphas,  Groups.chain=out.Groups, Groups=groups, groups_distribution=groups_distribution, accept.rate.tree=accept.rate.tree, accept.rate.tree.split=accept.rate.tree.split, accept.rate.tree.merge=accept.rate.tree.merge, accept.rate.alphas = accept.alphas.rate, eps = eps, alphas = (alphas[which(!is.na(alphas))]), sd = (sd[which(!is.na(sd))]), eps=eps,  burnin=burnin, eta=eta))
}

############
### MCMC ###
############
### Performs MCMC algorithm using the recursive formula of the nested logistic likelihood
# data: numeric matrix containing the extreme observations in Fréchet scale
# D: integer, data dimension
# R: number of iterations
# init.Groups: integer vector containing initial clustering configuration, same integer indicate that the corresponding variables belong to the same group
# init.D_K: integer vector containing the size of each initial group
# init.alphas.a: numeric, contains the initial value for the between clusters dependence parameter
# init.alphas.b: numeric vector, contains the initial value for the within clusters dependence parameter
# eps.init: adaptive tuning parameter for the proposal in the UPDATE move
# eta: arbitrary constant for the proposal in the SPLIT and MERGE moves, in practice should be chosen in order to ensure the RJ MCMC chain is well-mixing. We reccomend fixing it to 0.4.
MCMC <- function(data=na.omit(MAX), D=4, R = 1000, init.Groups =c(1,1,1,2), burnin=R/5, init.D_K=c(3,1), init.alphas.a = c(0.8), init.alphas.b = c(0.4, 0.8), eps.init = c(0.1, 0.1, 0.1), eta=0.4) { # eta too small means not much merging
    # need to think about how to define initial values for the parameters and tree structure
    n <- nrow(data) # sample size
    eps=eps.init
    K.max = D # there can be as many group as the dimensions, would be nice to make K.max flexible, i.e., K.max<=D
    out.K <- c() # chain for number of groups
    K<-out.K[1] <- length(init.alphas.b) # number of groups
    out.alphas.a <- c() # chain for alpha parameters
    out.alphas.a[1] <- init.alphas.a
    out.alphas.b <- matrix(NA, (R+1), K.max) # chain for alphas.b parameters
    out.alphas.b[1, 1:K] <- init.alphas.b
    D_K <- matrix(NA, (R+1), K.max) # chain for D_K
    D_K[1,1:K]<-init.D_K
    Groups <- matrix(NA, (R+1), D)
    Groups[1,] <- init.Groups # ordered depending on the apperance of the group in the columns of the data matrix
    log.post <- c()
    log.post[1]  <- -NegLikelihoodnestedLogistic2(c(init.alphas.a,  init.alphas.b), data, init.Groups, init.D_K)
    accept.alphas <- matrix(, ncol = (K.max+1), nrow = R)
    accept.tree <- c()
    accept.tree.split <- c()
    accept.tree.merge <- c()
    accept.tree.swap<- c()
    for (r in 1:R) {# main loop
        K=out.K[r]
        alphas.b.current=out.alphas.b[r, 1:K]
        Groups.current=Groups[r,]
        D_K.current=D_K[r, 1:K]
        eps.b=eps[2:(K+1)]
        UPDATE<-update.groups(alphas.b.current, Groups.current, D_K.current, eps.b, K, K.max)
        K<-UPDATE$K.new
        eps<-c(eps[1], UPDATE$eps.b)
        D_K[r, 1:K]<-UPDATE$D_K.new
        out.alphas.b[r, 1:K]<-UPDATE$alphas.b.new
        Groups[r,]<-UPDATE$Groups.new
        out.K[r]<-K
        Groups.current<-Groups[r,]
        D_K.current<-D_K[r, 1:K]
            ###############
            ### UPDATE ###
            ###############
            alphas.current<-c(out.alphas.a[r], out.alphas.b[r, 1:K])
            Groups[r+1,] <- Groups.current
            D_K[r+1, 1:K] <- D_K.current
            out.K[r+1]<-K
            log.post.current<-log.post[r]
            accept.alphas.current <- c()
            D_K.current.ab<-c(K, D_K.current) # we only care about the k group so the alpha parameter has k elements but doesn't matter
            for (p in 1:(K+1)) { #Loop over the parameters to estimate
                #if (D_K.current.ab[p]==1){
                #	alphas.current <- alphas.current
                #	log.post.current  <- log.post.current
                #	accept.alphas.current [p]<- NA
                #}else{
                if ((alphas.current[p] != 1) & (alphas.current[p] - eps[p]) < 1 & (alphas.current[p] + eps[p]) >= 1 & K!=K.max) { # If 1 is in the inverval but alpha_b is not exactly 1
                    # indicator for proposal distribution
                    if (runif(1) < 0.5) {
                        pr = 1
                        alpha.star <- runif(1, min = max(0.1, alphas.current[p] - eps[p]), max = min(1, alphas.current[p] + eps[p]))
                    } else {
                        pr = 2
                        alpha.star <- 1 # Point mass at 1 with probability 0.5
                    }
                } else {
                    pr = 0
                    alpha.star <- runif(1, min = max(0.1, alphas.current[p] - eps[p]), max = min(1, alphas.current[p] + eps[p])) ## generation of a new parameter centred around the current parameter (bouded at 0.1 and 1)
                }
                alphas.star <- alphas.current
                alphas.star[p] <- alpha.star
                log.post.star <- -NegLikelihoodnestedLogistic2(alphas.star, data, Groups.current, D_K.current)
                if (alpha.star == 1) { #the computation of the log probabilities changes depending from which proposal the candidate was drawn from
                    p_alpha_star=0.5+ 0.5*dunif(alpha.star, min = max(0.1, alphas.current[p] - eps[p]), max = min(1, alphas.current[p] + eps[p]), log = TRUE)
                }else{
                    if ((alphas.current[p] - eps[p]) < 1 & (alphas.current[p] + eps[p]) >= 1 & K!=K.max) {
                        p_alpha_star=0.5*dunif(alpha.star, min = max(0.1, alphas.current[p] - eps[p]), max = min(1, alphas.current[p] + eps[p]), log = TRUE)
                    }else{
                        p_alpha_star=dunif(alpha.star, min = max(0.1, alphas.current[p] - eps[p]), max = min(1, alphas.current[p] + eps[p]), log = TRUE)
                    }
                }
                if (alphas.current[p] == 1) { # the computation of the log probabilities changes depending from which proposal the candidate was drawn from
                    p_alpha_current=0.5+ 0.5*dunif(alphas.current[p], min = max(0.1, alpha.star - eps[p]), max = min(1, alpha.star + eps[p]), log = TRUE)
                }else{
                    if ((alpha.star - eps[p]) < 1 & (alpha.star + eps[p]) >= 1 & K!=K.max) {
                        p_alpha_current=0.5*dunif(alphas.current[p], min = max(0.1, alpha.star - eps[p]), max = min(1, alpha.star + eps[p]), log = TRUE)
                    }else{
                        p_alpha_current=dunif(alphas.current[p], min = max(0.1, alpha.star - eps[p]), max = min(1, alpha.star + eps[p]), log = TRUE)
                    }
                }
                log.prob <- min(0, log.post.star - log.post.current + p_alpha_star -  p_alpha_current) ## calculation of the log-acceptance probability
                if ((runif(1) < exp(log.prob)) & (alphas.current[p] != alpha.star)) { ## with probability exp(log.prob), we accept the new parameter, and update the chain accordingly, if the previous parameter value is the same as the candidate it is not count as an acceptance (suggested by Segers)
                    alphas.current <- alphas.star
                    log.post.current  <- log.post.star
                    accept.alphas[r, p]<- 1
                } else { ## with probability 1-exp(log.prob), we stay at the current parameter value, and update the chain accordingly
                    alphas.current <- alphas.current
                    log.post.current  <- log.post.current
                    accept.alphas[r, p]<- 0
                }
                if ((D_K.current.ab[p]>1) & (eps[p] + 0.05 < 0.2) & (mean(na.omit(accept.alphas[, p])) > 0.4)) { # Adaptive steps for epsilon (suggested by Emric)
                    eps[p] <- eps[p] + 0.05
                } else {
                    if ((D_K.current.ab[p]>1) & (eps[p] - 0.05 > 0.05) & (mean(na.omit(accept.alphas[, p])) < 0.2)) {
                        eps[p] <- eps[p] - 0.05
                    }
                }
                #}
            }
            out.alphas.a[r + 1] <- alphas.current[1]
            out.alphas.b[r + 1, 1:K]<-alphas.current[-1]
            log.post[r + 1] <- log.post.current
    } # end main loop
    chain<-cbind(out.alphas.a, out.alphas.b[1:(R+1),])
    #burnin <- floor(R/4)
    out.chain <- chain[burnin:(R+1), ] #Eliminate first observations
    out.Groups <- Groups[burnin:(R+1), ] 
    groups_distribution<-aggregate(list(numdup=rep(1,nrow(as.data.frame(out.Groups)))), as.data.frame(out.Groups), length) # computes the distribution of	
    groups <- groups_distribution[which.max(groups_distribution[, D+1]), 1:D] # Mode of Groups out of chain
    out.accept.alphas<-accept.alphas[burnin:R, ]
    out.alphas <- c()
    out.accept.alphas.rate<-c()
    for (i in 1:(R-burnin)) { #We only consider the chain values at spaced intervals for the estimation of the parameters 
        if (all(groups==out.Groups[i, ])){
            out.alphas <- rbind(out.alphas, out.chain[i, ])
            out.accept.alphas.rate <- rbind(out.accept.alphas.rate, out.accept.alphas[i, ])
        }
    }
    #thining <- as.numeric(round((nrow(out.alphas)/5), 0) - 1)
    #a = 0
    #out.al <- matrix(, ncol = ncol(chain), nrow = thining)
    #for (i in 1:thining) { #We only conder the chain values at spaced intervals for the estimation of the parameters 
        #a <- a + 5
        #out.al[i, ] <- out.alphas[a, ]
    #}
    out.al=out.alphas #?
    alphas <- apply(out.al, 2, function(x) median(na.omit(x))) # Mean of parameters estimates out of thinning chain
    sd <- (apply(out.al, 2, function(x) sd(na.omit(x)))/nrow(out.alphas)) #Sd estimates
    accept.rate.tree <- mean(na.omit(accept.tree)) #
    accept.rate.tree.split <- mean(na.omit(accept.tree.split)) #
    accept.rate.tree.merge <- mean(na.omit(accept.tree.merge)) #
    accept.rate.tree.swap <- mean(na.omit(accept.tree.swap)) #
    accept.alphas.rate = apply(out.accept.alphas.rate, 2, function(x) mean(na.omit(x)))
    return(list(chain = out.chain, out.alphas.chain = out.alphas,  Groups.chain=out.Groups, Groups=groups, groups_distribution=groups_distribution, accept.rate.tree=accept.rate.tree, accept.rate.tree.split=accept.rate.tree.split, accept.rate.tree.merge=accept.rate.tree.merge, accept.rate.alphas = accept.alphas.rate, eps = eps, alphas = (alphas[which(!is.na(alphas))]), sd = (sd[which(!is.na(sd))]), eps=eps,  burnin=burnin, eta=eta))
}


