// [[Rcpp::depends(RcppArmadillo)]]
#include <RcppArmadillo.h>
#include <Rcpp.h>
#include <vector>

using namespace Rcpp;
using namespace std;
using namespace sugar;
using namespace arma;


class Offset{
private:
    int nrows, ncols, nmats ;
    
public:
    Offset( IntegerVector dim ) : nrows(dim[0]), ncols(dim[1]),
    nmats(dim[2]){}
    
    Offset( int nrows_, int ncols_, int nmats_) : nrows(nrows_),
    ncols(ncols_), nmats(nmats_){}
    
    int operator()( int i, int j, int k){
        return i + j * nrows + k * ( nrows * ncols ) ;
    }
    
} ;

class Array3 : public NumericVector {
private:
    Offset offset ;
    
public:
    Array3( SEXP x) : NumericVector(x), offset(
                                               (IntegerVector)((RObject)x).attr("dim") ) { }
    Array3( Dimension dim ): NumericVector( dim ), offset( dim[0],
                                                          dim[1], dim[2] ) {}
    Array3( int nrows_, int ncols_, int nmats_ ): NumericVector(
                                                                Dimension(nrows_, ncols_, nmats_ ) ), offset( nrows_, ncols_, nmats_ ) {}
    
    inline double& operator()( int i, int j, int k){
        return ( (NumericVector*)(this) )->operator[]( offset( i, j, k)
                                                      ) ;
    }
    
} ;

/*##################################################################################################
 ### Full negative log likelihood function of the nested logistic model for componentwise maxima ###
 ###################################################################################################
 This computes the exponent measure for the D-dimensional nested logistic model
 Input
 *al: double, value of the alpha parameter
 *V_0: vector of size D, values of the V_1...V_K functions of the G groups
 Output
 *V: double, exponent measure V for the D-dimensional nested logistic model
 */
// [[Rcpp::export]]
double Vnl(double al, NumericVector V_0){
    int G=V_0.size();
    double Vk,V;
    double sum=0.0;
    int k;
    
    for (k=0; k<G; ++k){
        Vk=V_0[k];
        sum+=Vk;
    }
    V=R_pow(sum,al);
    return V;
}



/* This computes the vector exponent measures
 Input
 *al: double, contains the value of the alpha parameter
 *al_G: vector of doubles of size G (# of groups), contains the values of the parameters alpha_1, ..., alpha_G
 *Z: contains the value of the observations
 *Groups: vector of integers of size D, same integer indicate that the corresponding variables belong to the same group
 Output
 *v: vector of size K+1 containing V, V_1, ..., V_K
 */
// [[Rcpp::export]]
NumericVector VNestedLogistic(double al, NumericVector al_G, NumericVector Z, IntegerVector Groups){
    const int D=Z.size();
    int k, j, grp, p;
    double Zj, sum=0, a, V_0;
    IntegerVector Groups_unique=unique(Groups);
    Groups_unique=Groups_unique.sort();
    int G=Groups_unique.size();
    NumericVector v(G);
    for(k=0; k<G; ++k){
        grp=Groups_unique(k);
        a=al*al_G(k);
        sum=0;
        p=0;
        for(j=0; j<D; ++j){
            if(grp==Groups[j]){
                p+=1;
                Zj=Z[j];
                sum+=R_pow(Zj,-1.0/(a));
            }
        }
        v(k)=R_pow(sum,al_G(k));
    }
    V_0=Vnl(al, v);
    v.push_front(V_0);
    return v;
}



/* This computes a list containing the vector of exponent measures and part of the log likelihood function
 Input
 *al: double, value of the alpha parameter
 *al_k: double, value of the alpha_k parameter for the kth group
 *Z: vector of double of size D, observations
 *Groups: vector of integers of size D, same integer indicate that the corresponding variables belong to the same group
 Output 
 List of:
 *V: vector of double of size K, exponent measures v=V_0, V_1, ..., V_G
 *sum1: costant, sum_k sum_d (-1.0/(al*al_k)-1)*log(z_kd)
 */

// [[Rcpp::export]]
List V_prodNL(double al, NumericVector al_K, NumericVector Z, IntegerVector Groups){
    const int D=Z.size();
    int k, j, grp;
    long double Zj, sum=0, a, V_0, sum_z, c, sum1=0;
    List V_prod;
    IntegerVector Groups_unique=unique(Groups);
    Groups_unique=Groups_unique.sort();
    int K=Groups_unique.size();
    NumericVector V(K);

    for(k=0; k<K; ++k){
        grp=Groups_unique(k);
        a=al*al_K(k);
        sum=0;
        sum_z=0;
        for(j=0; j<D; ++j){
            if(grp==Groups[j]){
                Zj=Z[j];
                sum+=R_pow(Zj,-1.0/(a));
                sum_z+=log(Zj);
            }
        }
        V(k)=R_pow(sum,al_K(k));
        c=-(1.0/a);
        sum1+=(c-1)*sum_z;
    }
    V_0=Vnl(al, V);
    V.push_front(V_0);
    
    V_prod.push_back(V);
    
    V_prod.push_back(sum1);
    return V_prod;
}



/*
 This function determines the equivalent subscript values corresponding
 to the absolute index dimension of a multidimensional array.
Input
 *k: integer, index
 *D_K: integer vector containing the groups dimension
 Output
 *idx: vector containing the j1,...,j_K corresponding to the idx k
 
Example for a 2-D array of size [3, 5] : siz[] = {3, 5}; N = 2
 - if linear index is 0, subscript is (0, 0)
 - if linear index is 1, subscript is (0, 1)
 - if linear index is 5, subscript is (1, 0)
 - if linear index is 14, subscript is (2, 4)
 */

// [[Rcpp::export]]
IntegerVector index(int k, IntegerVector D_K){
    int l,j,prod=1;
    const int n=D_K.size();
    IntegerVector idx(n);
    IntegerVector r(n);
    IntegerVector prod1;
    if (n==1) {
        idx=k%(D_K[0]);
    }
    else
    {
        for (l=(n-1);l>0;l--){
            prod=prod*D_K[l];
            prod1.push_front(prod);
        }
        
        idx[0] = k / prod1[0];
        r[0]=0;
        r[1] = k % prod1[0];
        
        for (j=1;j<(n-1);++j){
            idx[j] = r[j] / prod1[j];
            r[j+1] = r[j] % prod1[j];
        }
        idx[(n-1)]= r[(n-1)];
    }
    return idx;
}

/*
This function determines the equivalent index value corresponding
to the subscript values of a multidimensional array.
*D_K: integer vector containing the groups dimension
*D: the data dimension
*sub: vector containing the j1,...,j_K
Output
*idx: the linear index corresponding to the subscript j1,...,j_K
 
Example for a 2-D array of size [3, 5] : siz[] = {3, 5}; N = 2
 - if subscript is (0, 0), linear index is 0
 - if subscript is (0, 1), linear index is 1
 - if subscript is (1, 0), linear index is 5
 - if subscript is (2, 4), linear index is 14
*/

// [[Rcpp::export]]
int invindex(IntegerVector D_K, int D, IntegerVector sub)
{
    int idx =	0;
    if (0)
        idx =		sub[0]*D_K[3]*D_K[2]*D_K[1]
        +	sub[1]*D_K[3]*D_K[2]
        +	sub[2]*D_K[3]
        +	sub[3];
    else
    {
        for (int i = 0; i < D; i++)
        {
            int prod = 1;
            for (int j = D-1; j > i; j--)
                prod *= D_K[j];
            idx += sub[i] * prod;
        }
    }
    return idx;
}


/*
 This function determines the value of the coefficients Beta for the nested logistic model (see formula pag.1 pdf)
Input
 *al: costant, containing the value of the parameter alpha
 *al_K: nueric vector of size K, containing the value of the parameters alpha_k, k=1:K
 *D: data dimension
 *K: number of groups
 *D_K: vector of size K, containing the number of obs per each group
 Output
 *List of coefficient Beta
 */

// [[Rcpp::export]]
List coeff_Beta(double al, NumericVector al_K, const int D, const int K, IntegerVector D_K){
    int k, KD, p, g, k_prod, j_prod, f, j_p, j_p_pr, j_p_1, j_p_11, c, sumj, l, n, sum_j;
    
    List Beta_K;
  //  IntegerMatrix sumj_vector(D_K[0], j_prod);

    for (k=0;k<K;++k){ // loop over the groups
        sum_j=0;
        j_prod=1;
        for(g=0;g<k+1;g++) {
            sum_j+=D_K[g]; //max j_1+...+j_k
            j_prod=j_prod*D_K[g]; //max j_1*...*j_k
        }
        Array3 Beta(j_prod, sum_j, D_K[k]); //j,l,D_k
        
        IntegerVector kD_K_pr(k);
        IntegerVector kD_K_1(k+1);
        IntegerVector kD_K(k+1); // k+1 is the # of groups considered so far in the loop
        
        for (KD=1;KD<=D_K[k];++KD){ //loop over the D_k in order to derive recursive coeff D_k+1
            // loop over j_1*...*j_k considered so far
            k_prod=1;
            for (p=0;p<k;++p){
                k_prod=k_prod*D_K[p];
                kD_K_pr[p]=D_K[p];
                kD_K_1[p]=D_K[p];
                kD_K[p]=D_K[p];
            }
            k_prod=k_prod*KD;
            kD_K_1[k]=KD-1;
            kD_K[k]=KD;
            
            for (j_p=0;j_p<k_prod;++j_p){
                IntegerVector j0=index(j_p, kD_K);
                IntegerVector j=j0+1; // not sure this add 1 to all elements, check!
                n=j.size();
                IntegerVector j_pr(n-1);
                IntegerVector j1(n);
                for (f=0;f<k;++f){
                    j_pr[f]=j0[f];
                    j1[f]=j0[f];
                }
                j1[k]=j0[k]-1;
                j_p_pr=invindex(kD_K_pr, (n-1), j_pr);
                j_p_1=invindex(kD_K_1, n, j0);
                j_p_11=invindex(kD_K_1, n, j1);
                // need to compute the sum of the js
                
                sumj=0;
                for (c=0;c<n;++c){
                    sumj+=j[c]; // sumj also in the scale of the js, KD
                }
                
             //   sumj_vector(KD-1, j_p)=sumj;
                // find j_p-1
                for (l=1;l<=sumj;++l){ // loop over the l's
                    // two cases:
                    // first case new group
                    if(KD==1){
                        if(k==0) {// in this case conditions are only on the l's
                            Beta(j_p, l-1, KD-1)=1;
                            
                        }
                        else if (l==1){
                        Beta(j_p, l-1, KD-1)=as<Array3>(Beta_K[k-1])(j_p_pr, l-1, D_K[k-1]-1)*(-1)*(l-(sumj-1)/al); //j_p are the same because the last j_k+1 is equal to 1 so doesn't change in the multiplication
                            
                            
                        }
                        else if (l==sumj) {
                            Beta(j_p, l-1, KD-1)=as<Array3>(Beta_K[k-1])(j_p_pr, l-2, D_K[k-1]-1);
                           
                        }
                        else {
                         Beta(j_p, l-1, KD-1)=as<Array3>(Beta_K[k-1])(j_p_pr, l-1, D_K[k-1]-1)*(-1)*(l-(sumj-1)/al)+as<Array3>(Beta_K[k-1])(j_p_pr, l-2, D_K[k-1]-1);
                            
                        }
                    }
                    // second case add elements to a group, here the conditions are on the j[k]=j_k+1 and on the l's
                    else {
                        if (j[k]==1 && 1<=l<=sumj){
                        Beta(j_p, l-1, KD-1)=Beta(j_p_1, l-1, KD-2)*(j[k]-((KD-1)/al_K[k]))*(-1/al);
                           
                        }
                        else if (j[k]==KD && l==sumj){
                            Beta(j_p, l-1, KD-1)=Beta(j_p_11, l-2, KD-2);
                           

                        }
                        else if(j[k]==KD && l==1){
                            Beta(j_p, l-1, KD-1)=Beta(j_p_11, l-1, KD-2)*(-1)*(l-((sumj-1)/al));
                            
                        }
                        else if (2<=j[k]<=KD-1 && l==1){
                            Beta(j_p, l-1, KD-1)=Beta(j_p_11, l-1, KD-2)*(-1)*(l-((sumj-1)/al))+Beta(j_p_1, l-1, KD-2)*(j[k]-((KD-1)/al_K[k]))*(-1/al);
                            
                        }
                        else if (j[k]==KD && 2<=l<=sumj-1){
                            Beta(j_p, l-1,KD-1)=Beta(j_p_11, l-2, KD-2)+Beta(j_p_11, l-1, KD-2)*(-1)*(l-((sumj-1)/al));
                            
                        }
                        else if(2<=j[k]<=KD-1 && 2<=l<=sumj){
                            Beta(j_p, l-1, KD-1)=Beta(j_p_11, l-2, KD-2)+Beta(j_p_11, l-1, KD-2)*(-1)*(l-((sumj-1)/al))+Beta(j_p_1, l-1, KD-2)*(j[k]-((KD-1)/al_K[k]))*(-1/al);
                        }
                        else{
                            Beta(j_p, l-1, KD-1)=-1;
                        }
                    }
                }
            }
        }
        Beta_K.push_back(Beta);
    }
  return Beta_K;
}


/*
 This computes the product of the V_k
 Input
 *al_K: nueric vector of size K, containing the value of the parameters alpha_k, k=1:K
 *K: number of groups
 *v: vector of size K+1 containing V, V_1, ..., V_K
 *j: vector containing the j1,...,j_K corresponding to the idx k
 *D_K: vector of size K, containing the number of obs per each group
 Output
 *prod_v: double, product of the V_k
 */

// [[Rcpp::export]]
 double Prod(NumericVector al_K, int K, NumericVector v, IntegerVector j, IntegerVector D_K){
    double v_k, prod_v=1, prod, f;
    for (int p=0;p<K;++p){
        v_k=v[p+1];
        f=(j[p]-(D_K[p]/al_K[p]));
        prod=R_pow(v_k,f);
        prod_v=prod_v*prod;
    }
    return prod_v;
}


/*
 This computes the full negative log-likelihood for the D-dimensional nested logistic model
Input
*alphas: vector of size K+1, containing the value of the parameters alpha and alpha_k, k=1:K
*obs: matrix of size nxd, containing the data (each row= 1 observed vector)
*Group: vector of size d, containing the group index
*Beta: matrix of coefficients computed using coeff_Beta()
*D_K: vector of size K, containing the number of obs per each group
Output 
*-sum: double, negative log-likelihood
*/
// [[Rcpp::export]]
double NegLikelihoodnestedLogistic(NumericVector alphas, NumericMatrix obs, IntegerVector Groups, IntegerVector D_K){
    double sum=0.000;
    double p, b, Q, v_0, g, sum1, al=alphas[0],r;
    const int n=obs.nrow();
    const int D=obs.ncol();
    List v_prod;
    List Beta_K;
    int i, k, j_p, c, k_prod=1, sumj, l;
  
    IntegerVector Groups_unique=unique(Groups);
    int K=Groups_unique.size();
    NumericVector al_K(K);
    Groups_unique=Groups_unique.sort();
    for (k=0;k<K;++k){
        k_prod=k_prod*D_K[k];
        al_K[k]=(alphas[k+1]);
    }
    
    Beta_K=coeff_Beta(al, al_K, D, K, D_K);
    Array3 Beta=as<Array3>(Beta_K[K-1]);
    
    NumericVector v(K+1);
    for (i=0;i<n;++i){ //loop over replicates
        v_prod=V_prodNL(al,al_K,obs(i,_),Groups);
        r=as<double>(v_prod[1]);
        sum+=r;
        v=as<NumericVector>(v_prod[0]);
        sum+=-v[0];
        sum1=0.000;
        for (j_p=0;j_p<k_prod;++j_p){
            IntegerVector j0=index(j_p, D_K);
            IntegerVector j=j0+1;
            sumj=0;
            for (c=0;c<K;++c){
                sumj+=j[c]; // sumj also in the scale of the js, KD
            }
            for (l=1;l<=sumj;++l){ // loop over the l's
                b=Beta(j_p, l-1,(D_K[K-1]-1)); //positive and negative
                g=(l-(sumj/al));
                v_0=R_pow(v[0], g);
                p=Prod(al_K, K, v, j, D_K);
                Q=p*v_0*b;
                sum1+=Q;
            }
        }
        sum+=log(sum1);
    }
    return -sum;
}

/*
 This computes the log of the product of the V_k
 Input
 *al_K: nueric vector of size K, containing the value of the parameters alpha_k, k=1:K
 *K: number of groups
 *v: vector of size K+1 containing V, V_1, ..., V_K
 *j: vector containing the j1,...,j_K corresponding to the idx k
 *D_K: vector of size K, containing the number of obs per each group
 Output
 *prod_v: double, log of the product of the V_k
*/

// [[Rcpp::export]]
long double Prod2(NumericVector al_K, int K, NumericVector v, IntegerVector j, IntegerVector D_K){
    long double v_k, prod_v=0, prod, f;
    for (int p=0;p<K;++p){
        v_k=v[p+1];
        f=(j[p]-(D_K[p]/al_K[p]));
        prod=f*log(v_k);
        prod_v+=prod;
    }
    return prod_v;
}

/*
 This computes the full negative log-likelihood for the D-dimensional nested logistic model, using the log for the product inside the sums
 Input
 *alphas: vector of size K+1, containing the value of the parameters alpha and alpha_k, k=1:K
 *obs: matrix of size nxd, containing the data (each row= 1 observed vector)
 *Group: vector of size d, containing the group index
 *Beta: matrix of coefficients computed using coeff_Beta()
 *D_K: vector of size K, containing the number of obs per each group
 Output
 *-sum: double, negative log-likelihood
*/
// [[Rcpp::export]]
double NegLikelihoodnestedLogistic2(NumericVector alphas, NumericMatrix obs, IntegerVector Groups, IntegerVector D_K){
    long double sum=0.000;
    long double r, b, Q, p, v_0, g, sum1, al=alphas[0];
    const int n=obs.nrow();
    const int D=obs.ncol();
    List v_prod;
    List Beta_K;
    int i, k, j_p, c, k_prod=1, sumj, l;
    if (al>1| al<=0) {
        return 1000000000000000;
    }
    IntegerVector Groups_unique=unique(Groups);
    int K=Groups_unique.size();
    NumericVector al_K(K);
    Groups_unique=Groups_unique.sort();
    for (k=0;k<K;++k){
        k_prod=k_prod*D_K[k];
        al_K[k]=(alphas[k+1]);
        if (al_K[k]>1| al_K[k]<=0) {
            return 1000000000000000;
        }
    }
    NumericVector v(K+1);
    Beta_K=coeff_Beta(al, al_K, D, K, D_K);
    Array3 Beta=as<Array3>(Beta_K[K-1]);

    for (i=0;i<n;++i){ //loop over replicates
        v_prod=V_prodNL(al,al_K,obs(i,_),Groups);
        v=as<NumericVector>(v_prod[0]);
        sum+=-v[0];
        //Rprintf("v[0]=%f\n",-v[0]); // i for integer
        r=as<double>(v_prod[1]);
        sum+=r;
        //Rprintf("r=%f\n",r); // i for integer

        sum1=0;
        for (j_p=0;j_p<k_prod;++j_p){
            IntegerVector j0=index(j_p, D_K);
            IntegerVector j=j0+1;
            sumj=0;
            for (c=0;c<K;++c){
                sumj+=j[c]; // sumj also in the scale of the js, KD
            }
            for (l=1;l<=sumj;++l){ // loop over the l's
                b=log(Beta(j_p, l-1,(D_K[K-1]-1))); //positive and negative
                g=(l-(sumj/al));
                v_0=g*log(v[0]);
                p=Prod2(al_K, K, v, j, D_K);
                Q=p+v_0+b;
                sum1+=exp(Q);
            }
        }
        sum+=log(sum1);
        
    }
    return -sum;
}



